import json
import time
from functools import wraps

from errors import Error
from flask import abort, current_app, make_response, request, session, Response
import _globals
from _token import Token
import util
import validators



# Make a JSON response out of the dict that's passed in
def to_json_response(d, code=200):
   jd = json.dumps(d)
   return make_response(jd, code, { 'Content-Type': 'application/json' })

# @render_as_json
# 1. If the service is restarting, return an error message
# 2. If a payload is present for PUT/POST methods, ensure that it's JSON
# 3. Render the returning dict as JSON
def render_as_json(f):
   def wrapped_f(*args, **kwargs):
      if _globals.restarting:
         return to_json_response(Error.RESTARTING, 503)
      methods = request.url_rule.methods
      if _globals.upgrading and (not request.path.startswith('/api/v1/update')
                                 or 'GET' not in methods):
         # while the system is installing, only querying the update page is
         # allowed.
         return to_json_response(Error.INSTALLING, 503)
      if "PUT" in methods or "POST" in methods:
         if request.content_length > 0 and len(request.files) == 0:
            data = request.get_json(silent=True)
            if request.mimetype != 'application/json' or not data:
               return to_json_response(Error.MALFORMED_DATA, 400)
      try:
         rv = f(*args, **kwargs)
         if rv != None:
            if isinstance(rv, Response):
               return rv
            msg, code = rv if isinstance(rv, (list, tuple)) else (rv, 200)
            return to_json_response(msg, code)
      except validators.ValidationError as e:
         return to_json_response(e.to_error(), 400)
   wrapped_f.__name__ = f.__name__
   return wrapped_f

# @authenticate
# Authenticate the API request and return a JSON or HTTP response. In addition
# we also check if first_login is True. In that case, force the user to change
# the default password before s/he can continue.
# Supports authentication methods:
# * token (generated by login)
# * session cookie that contains the token
# * valid client certificate
def authenticate(html=False, chpass=True, cert_required=False):
   def wrap(f):
      def wrapped_f(*args, **kwargs):

         def make_reply(err_code, status, text):
            return make_response(text, status) if html else (err_code, status)

         token = request.headers.get('X-API-Token')
         client_cn = request.headers.get('X-SSL-Client-CN')
         if cert_required and not client_cn:
            return (Error.CERT_REQUIRED, 401)
         if not token:
            token = session.get('token')
         if not token and client_cn:
            return f(*args, **kwargs)
         if not Token.is_valid(current_app, token):
            return make_reply(Error.AUTH_FAILED, 401, "Authentication Failed")
         if util.is_first_login() and chpass:
            return make_reply(Error.CHANGE_PASSWORD, 200, "Change Default Password")
         return f(*args, **kwargs)
      wrapped_f.__name__ = f.__name__
      return wrapped_f
   return wrap

# @restart_service
# Set the global restarting flag so that subsequent API requests will fail,
# until the dashboard service is restarted
def restart_service(f):
   def wrapped_f(*args, **kwargs):
      rval = f(*args, **kwargs)
      _globals.restarting = True
      return rval
   wrapped_f.__name__ = f.__name__
   return wrapped_f


# @throttle(seconds)
# Invoke the decorated function once every 'seconds' seconds, even if the
# function is called multiple times. This also adds a 'force' method to the
# original function to bypass the throttle check and invoke the underlying
# function.
class throttle(object):
   def __init__(self, seconds):
      self._throttle_period = seconds
      self._time_of_last_call = time.time()

   def __call__(self, fn):
      def force(*args, **kwargs):
         return fn(*args, **kwargs)
      fn.force = force
      @wraps(fn)
      def wrapper(*args, **kwargs):
         now_time = time.time()
         elapsed_time = now_time - self._time_of_last_call
         if elapsed_time > self._throttle_period:
            self._time_of_last_call = now_time
            return fn(*args, **kwargs)
      return wrapper

# @manager_route
# Route is available only in centralized manager mode
def manager_route(f):
   @wraps(f)
   def wrapper(*args, **kwargs):
      if util.is_cmr():
         return f(*args, **kwargs)
      abort(404)
   return wrapper

# @appliance_only_route
# Route is blocked in centralized manager mode
def appliance_only_route(f):
   @wraps(f)
   def wrapper(*args, **kwargs):
      if not util.is_primary_cmr():
         return f(*args, **kwargs)
      abort(404)
   return wrapper
